#!/usr/bin/env python
# -*- coding: utf-8 -*-
import requests

from . import helper

# some "consts"
V2_ENTITIES = "/v2/entities/"
LD_ENTITIES = "/ngsi-ld/v1/entities/"
OPTIONS_UPSERT = "?options=upsert"
ATTRS = "/attrs/"
TYPE_IS = "?type="


def do_delete(session, host, ngsi_id, headers, args):
    if args.protocol == helper.PROTOCOL_NGSI_V2:
        url = host + V2_ENTITIES
    else:
        url = host + LD_ENTITIES

    try:
        url_delete = url + helper.create_id(
            ngsi_id,
            args.prefix,
            args.postfix,
            0,
            args.protocol == helper.PROTOCOL_NGSI_LD,
        )
        return session.delete(url_delete, headers=headers)
    except (requests.exceptions.RequestException, Exception):
        return None


def do_post(session, host, first_id, headers, upsert, args):
    if args.protocol == helper.PROTOCOL_NGSI_V2:
        url = host + V2_ENTITIES
        if upsert:
            url += OPTIONS_UPSERT
        payload = helper.create_payload_ngsi_v2(first_id, True, args)
    else:
        url = host + LD_ENTITIES
        if upsert:
            url += OPTIONS_UPSERT
        payload = helper.create_payload_ngsi_ld(first_id, args, True)

    try:
        resp = session.post(url, data=payload, headers=headers)
    except (requests.exceptions.RequestException, Exception):
        return None, payload

    return resp, payload


def do_patch(session, host, first_id, headers, args):
    if args.protocol == helper.PROTOCOL_NGSI_V2:
        url = (
            host
            + V2_ENTITIES
            + helper.create_id(first_id, args.prefix, args.postfix, 0, False)
            + ATTRS
        )
        if args.type is not None:
            url += TYPE_IS
            url += args.type

        payload = helper.create_payload_ngsi_v2(None, False, args)
    else:
        url = (
            host
            + LD_ENTITIES
            + helper.create_id(first_id, args.prefix, args.postfix, 0, True)
            + ATTRS
        )
        payload = helper.create_payload_ngsi_ld(None, args, False)

    try:
        resp = session.patch(url, data=payload, headers=headers)
    except (requests.exceptions.RequestException, Exception):
        return None, payload

    return resp, payload



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import json
import socket
import sys
from time import sleep

import paho.mqtt.client as mqtt  # type: ignore
import requests

from . import helper

# some "const" values
INVALID_ID: int = -1
ERROR: int = -2


def get_sensor_things_relevant_attribute_count(args):
    num_attributes = 0

    if args.numbers is not None:
        num_attributes += len(args.numbers)
    if args.strings is not None:
        num_attributes += len(args.strings)

    return num_attributes


def init_mqtt(host):
    mqtt_client = mqtt.Client()

    try:
        mqtt_client.connect(host.split("/")[2].split(":")[0], 1883, 60)
        print("Waiting for MQTT-Broker to be connected...", end="", flush=True)
        mqtt_client.loop_start()

        while not mqtt_client.is_connected():
            print(".", end="", flush=True)
            sleep(0.1)

        print("OK", flush=True)
    except (socket.gaierror, Exception):
        print("\nError connecting to MQTT-Broker at '%s'!\nExiting..." % host)
        sys.exit(0)

    return mqtt_client


def delete_thing(session, host, thing_id, args):
    url = "%s/v1.1/Things(%i)" % (host, thing_id)

    headers = {helper.CONTENT_TYPE: helper.APPLICATION_JSON}

    if args.headers is not None:
        headers.update(args.headers)

    return session.delete(url, headers=headers)


def create_thing(session, host, thing_name, args):
    try:
        url = "%s/v1.1/Things" % host

        headers = {helper.CONTENT_TYPE: helper.APPLICATION_JSON}

        if args.headers is not None:
            headers.update(args.headers)

        payload = helper.create_thing_payload(thing_name, args.indent)

        resp = session.post(url, data=payload, headers=headers)

        if resp.status_code == 201:
            thing_id, ms2 = get_thing_id(session, host, thing_name, args)

            return thing_id, resp

    except (requests.exceptions.RequestException, Exception):
        return INVALID_ID, None

    return INVALID_ID, resp


def create_data_stream(session, host, thing_id, data_stream_name, args):
    try:
        url = "%s/v1.1/Things(%i)/Datastreams" % (host, thing_id)

        headers = {helper.CONTENT_TYPE: helper.APPLICATION_JSON}

        if args.headers is not None:
            headers.update(args.headers)

        payload = helper.create_data_stream_payload(data_stream_name, args.indent)

        resp = session.post(url, data=payload, headers=headers)
        ms = int(resp.elapsed.total_seconds() * 1000)

        if resp.status_code == 201:
            thing_id, ms2 = get_data_stream_id(
                session, host, thing_id, data_stream_name, args
            )

            ms = int((ms + ms2) / 2)
            return thing_id, ms

    except (requests.exceptions.RequestException, Exception):
        return ERROR, 0

    return INVALID_ID, ms


def create_observation(
    mqtt_client, session, host, use_mqtt, data_stream_id, value, args
):
    payload = helper.create_observation_payload(value, args.indent)

    if use_mqtt:
        topic = "v1.1/Datastreams(%i)/Observations" % data_stream_id
        mqtt_client.publish(topic, payload)
        # simulate http response
        resp = requests.Response()
        resp.status_code = 201
        return resp
    else:
        try:
            url = "%s/v1.1/Datastreams(%i)/Observations" % (host, data_stream_id)

            headers = {helper.CONTENT_TYPE: helper.APPLICATION_JSON}

            if args.headers is not None:
                headers.update(args.headers)

            return session.post(url, data=payload, headers=headers)

        except (requests.exceptions.RequestException, Exception):
            return None


def get_thing_id(session, host, thing_name, args):
    try:
        url = "%s/v1.1/Things?$select=name,id&$filter=name eq '%s'" % (host, thing_name)

        headers = {}

        if args.headers is not None:
            headers.update(args.headers)

        resp = session.get(url, headers=headers)

        if resp.status_code == 200:
            obj = json.loads(resp.text)
            value = obj["value"]

            if len(value) > 0:
                return value[0]["@iot.id"], resp
            else:
                # not found
                resp.status_code = 404
        else:
            return ERROR, resp
    except (requests.exceptions.RequestException, Exception):
        return ERROR, None

    return INVALID_ID, resp


def get_data_stream_id(session, host, thing_id, data_stream_name, args):
    try:
        url = "%s/v1.1/Things(%i)/Datastreams?$select=name,id&$filter=name eq '%s'" % (
            host,
            thing_id,
            data_stream_name,
        )

        headers = {}

        if args.headers is not None:
            headers.update(args.headers)

        resp = session.get(url, headers=headers)

        if resp.status_code == 200:
            obj = json.loads(resp.text)
            value = obj["value"]

            if len(value) > 0:
                return value[0]["@iot.id"], int(resp.elapsed.total_seconds() * 1000)
    except (requests.exceptions.RequestException, Exception):
        return ERROR, 0

    return INVALID_ID, 0



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import signal
import sys
from datetime import datetime
from threading import Lock, Thread
from time import sleep
from typing import Dict

import requests

from .modules import arguments, helper, ngsi, output, sensor_things

# some globals
start = datetime.now()
errors = 0
unique_errors = dict()  # type: Dict[str, int]
overall_messages = 0
overall_time = 0
deleted = 0
not_deleted = 0
send_threads = []
delete_thread = Thread()
halt = False


def do_delete(session, lock, args, max_id_length):
    global errors, deleted, not_deleted, overall_messages
    delimiter = ""

    host = helper.create_host_url(args.server)

    headers = {}

    if args.headers is not None:
        headers.update(args.headers)

    for i in range(args.delete[0], args.delete[1] + 1):
        if halt:
            return
        connection_error = False

        if (
            args.protocol == helper.PROTOCOL_NGSI_V2
            or args.protocol == helper.PROTOCOL_NGSI_LD
        ):
            resp = ngsi.do_delete(session, host, i, headers, args)
            if resp is None:
                connection_error = True
            else:
                if resp.status_code == 204:
                    deleted += 1
                elif resp.status_code == 404:
                    not_deleted += 1
                else:
                    errors += 1
        else:
            thing_name = helper.create_id(i, args.prefix, args.postfix, 0, False)
            thing_id, ms = sensor_things.get_thing_id(session, host, thing_name, args)
            if thing_id == sensor_things.INVALID_ID:
                not_deleted += 1
                resp = requests.Response()
                resp.status_code = 404
            else:
                resp = sensor_things.delete_thing(session, host, thing_id, args)
                if resp.status_code == 200:
                    deleted += 1
                else:
                    errors += 1

        with lock:
            if resp is None:
                errors += 1
                error_as_string = "Connection Error"
                if error_as_string in unique_errors.keys():
                    num = unique_errors[error_as_string]
                    unique_errors[error_as_string] = num + 1
                else:
                    unique_errors[error_as_string] = 1
            else:
                error_as_string = (
                    str(resp.status_code) + " " + " ".join(resp.text.split())[0:120]
                )

                if error_as_string in unique_errors.keys():
                    num = unique_errors[error_as_string]
                    unique_errors[error_as_string] = num + 1
                else:
                    unique_errors[error_as_string] = 1

            overall_messages += 1

            if args.verbose:
                if connection_error:
                    print(
                        "%s%s  ??? %s"
                        % (
                            delimiter,
                            helper.create_id(
                                i,
                                args.prefix,
                                args.postfix,
                                max_id_length,
                                args.protocol == helper.PROTOCOL_NGSI_LD,
                            ),
                            error_as_string,
                        ),
                        end="",
                    )
                else:
                    print(
                        "%s%s  %3i %s"
                        % (
                            delimiter,
                            helper.create_id(
                                i,
                                args.prefix,
                                args.postfix,
                                max_id_length,
                                args.protocol == helper.PROTOCOL_NGSI_LD,
                            ),
                            resp.status_code,
                            " ".join(resp.text.split())[0:120],
                        ),
                        end="",
                    )
                delimiter = "\n"


def do_send(mqtt_client, session, lock, args, offset, max_id_length):
    global errors, unique_errors, overall_time, overall_messages

    first_id = args.first_id + offset
    host = helper.create_host_url(args.server)

    # We'll start with unknown Thing/Datastream...let's see, if we can cache them later
    # thing_id will not change if --static-id is set!
    thing_id = sensor_things.INVALID_ID
    # data_stream_id will not change if only ONE attribute is set!
    data_stream_id = sensor_things.INVALID_ID

    if args.unlimited:
        # there is no infinite mode :-) - but don't send more than one billion messages!
        args.num_messages = 1000000000

    # how many attributes (relevant for SensorThings) do we have at all?
    num_attributes = sensor_things.get_sensor_things_relevant_attribute_count(args)

    for i in range(args.num_messages):
        start_time_ms = datetime.now()

        if halt:
            return

        if (
            args.protocol == helper.PROTOCOL_NGSI_V2
            or args.protocol == helper.PROTOCOL_NGSI_LD
        ):
            if args.protocol == helper.PROTOCOL_NGSI_V2:
                headers = {helper.CONTENT_TYPE: helper.APPLICATION_JSON}
            else:
                headers = {helper.CONTENT_TYPE: helper.APPLICATION_JSON_LD}

            if args.headers is not None:
                headers.update(args.headers)

            if args.insert_always:
                resp, payload = ngsi.do_post(
                    session, host, first_id, headers, True, args
                )
                if resp is None:
                    ms = 0
                    okay = False
                else:
                    ms = int(resp.elapsed.total_seconds() * 1000)
                    okay = resp.status_code == 204 or resp.status_code == 201
            else:
                resp, payload = ngsi.do_patch(session, host, first_id, headers, args)
                if resp is None:
                    ms = 0
                    okay = False
                else:
                    ms = int(resp.elapsed.total_seconds() * 1000)
                    if resp.status_code == 404:
                        # It would be very okay not to use "options=upsert" in this POST, but when
                        # running more than one thread, those POSTs will interfere each other!
                        # The 5th parameter is True in order to use "options=upsert"
                        # TODO: Orion-LD (currently) does not support upsert at all! As soon as
                        # this works, set 5th parameter to True again
                        resp, payload = ngsi.do_post(
                            session,
                            host,
                            first_id,
                            headers,
                            args.protocol == helper.PROTOCOL_NGSI_V2,
                            args,
                        )
                        if resp is None:
                            ms = 0
                        else:
                            ms += int(resp.elapsed.total_seconds() * 1000)
                    # At this point, resp.status_code has to be either 204 (after PATCH)
                    # or 201 (after POST) in order to be "okay"
                    if resp.status_code == 204 or resp.status_code == 201:
                        okay = True
                    else:
                        okay = False

            with lock:
                if not okay:
                    errors += 1
                    if resp is None:
                        error_as_string = "Connection Error"
                    else:
                        error_as_string = (
                            str(resp.status_code)
                            + " "
                            + " ".join(resp.text.split())[0:160]
                        )

                    if error_as_string in unique_errors.keys():
                        num = unique_errors[error_as_string]
                        unique_errors[error_as_string] = num + 1
                    else:
                        unique_errors[error_as_string] = 1

                overall_time += ms
                overall_messages += 1

                if args.verbose:
                    if resp is None:
                        print(
                            "%s  ???  ---- Connection Error!"
                            % helper.create_id(
                                first_id,
                                args.prefix,
                                args.postfix,
                                max_id_length,
                                args.protocol == helper.PROTOCOL_NGSI_LD,
                            )
                        )
                    else:
                        message = " ".join(resp.text.split())[0:120]
                        # There is this funny thing that Orion sometimes tells
                        # us about 400 ParseError
                        # see https://github.com/telefonicaid/fiware-orion/issues/3731
                        if resp.status_code == 400:
                            message += "\nPayload was:\n" + payload

                            # try to resend
                            # if insert_always:
                            #     url = host + "/v2/entities/?options=upsert"
                            #
                            #     resp = session.post(url, data=payload, headers=headers)

                            # In 100%, resending exactly the same payload again gives a 201

                        print(
                            "%s  %3i  %4i %s"
                            % (
                                helper.create_id(
                                    first_id,
                                    args.prefix,
                                    args.postfix,
                                    max_id_length,
                                    args.protocol == helper.PROTOCOL_NGSI_LD,
                                ),
                                resp.status_code,
                                ms,
                                message,
                            )
                        )
        else:
            #  Here we go with SensorThings-HTTP/SensorThings-MQTT
            resp = None
            ms = 0

            if args.datastream_id is not None:
                okay = True
                thing_id = 0
                data_stream_id = args.datastream_id
            else:
                if thing_id != sensor_things.INVALID_ID:
                    okay = True
                else:
                    #  1. Get the Thing's id and create if not existing yet
                    with lock:  # since we might be running in more than one thread, use lock!
                        okay = True

                        if thing_id == sensor_things.INVALID_ID:
                            thing_name = helper.create_id(
                                first_id,
                                args.prefix,
                                args.postfix,
                                0,
                                args.protocol == helper.PROTOCOL_NGSI_LD,
                            )

                            #  check, if the thing with the given name (thing_name) already exists:
                            thing_id, resp = sensor_things.get_thing_id(
                                session, host, thing_name, args
                            )
                            if resp is None:
                                okay = False
                            elif resp.status_code == 404:
                                ms = int(resp.elapsed.total_seconds() * 1000)
                                thing_id, resp = sensor_things.create_thing(
                                    session, host, thing_name, args
                                )
                                if resp.status_code != 201:
                                    okay = False
                            elif resp.status_code == 200:
                                # everything is fine
                                okay = True
                            else:
                                # neither 200 nor 404 - error!
                                okay = False

            if (
                okay
                and thing_id != sensor_things.INVALID_ID
                and thing_id != sensor_things.ERROR
            ):
                if args.numbers is not None:
                    for number in args.numbers:
                        attribute_args = number[0].split(",")

                        # shortcut: if at this point data_stream_id != INVALID_ID, then we
                        #           only have ONE attribute at all! So, no need to find out
                        #           data_stream_id...create an Observation immediately instead
                        if (
                            args.datastream_id is None
                            and data_stream_id == sensor_things.INVALID_ID
                        ):
                            with lock:  # since we might be running in more than one thread!
                                data_stream_name = attribute_args[0]
                                data_stream_id, ms2 = sensor_things.get_data_stream_id(
                                    session, host, thing_id, data_stream_name, args
                                )

                                ms = int((ms + ms2) / 2)

                                if data_stream_id == sensor_things.ERROR:
                                    okay = False
                                elif data_stream_id == sensor_things.INVALID_ID:
                                    (
                                        data_stream_id,
                                        ms2,
                                    ) = sensor_things.create_data_stream(
                                        session, host, thing_id, data_stream_name, args
                                    )
                                    ms = int((ms + ms2) / 2)
                                    if data_stream_id == sensor_things.ERROR:
                                        okay = False
                                        data_stream_id = sensor_things.INVALID_ID

                        # Create an observation
                        value = helper.create_value_from_attribute_args(attribute_args)

                        resp = sensor_things.create_observation(
                            mqtt_client,
                            session,
                            host,
                            args.protocol == helper.PROTOCOL_SENSOR_THINGS_MQTT,
                            data_stream_id,
                            value,
                            args,
                        )

                        if resp is not None:
                            ms2 = int(resp.elapsed.total_seconds() * 1000)
                            ms = int((ms + ms2) / 2)

                        if num_attributes > 1:
                            data_stream_id = sensor_things.INVALID_ID

                if args.strings is not None:
                    for string in args.strings:
                        data_stream_name = string[0]
                        with lock:  # since we might be running in more than one thread, use lock!
                            data_stream_id, ms2 = sensor_things.get_data_stream_id(
                                session, host, thing_id, data_stream_name, args
                            )

                            ms = int((ms + ms2) / 2)

                            if data_stream_id == sensor_things.ERROR:
                                okay = False
                            elif data_stream_id == sensor_things.INVALID_ID:
                                data_stream_id, ms2 = sensor_things.create_data_stream(
                                    session, host, thing_id, data_stream_name, args
                                )
                                ms = int((ms + ms2) / 2)
                                if data_stream_id == sensor_things.ERROR:
                                    okay = False
                                    data_stream_id = sensor_things.INVALID_ID

                        # Create an observation
                        resp = sensor_things.create_observation(
                            mqtt_client,
                            session,
                            host,
                            args.protocol == helper.PROTOCOL_SENSOR_THINGS_MQTT,
                            data_stream_id,
                            string[1],
                            args,
                        )

                        if resp is not None:
                            ms2 = int(resp.elapsed.total_seconds() * 1000)
                            ms = int((ms + ms2) / 2)

                        if num_attributes > 1:
                            data_stream_id = sensor_things.INVALID_ID

            with lock:
                overall_messages += 1

                if not okay:
                    errors += 1
                    if resp is None:
                        error_as_string = "Connection Error"
                    else:
                        error_as_string = (
                            str(resp.status_code)
                            + " "
                            + " ".join(resp.text.split())[0:120]
                        )

                    if error_as_string in unique_errors.keys():
                        num = unique_errors[error_as_string]
                        unique_errors[error_as_string] = num + 1
                    else:
                        unique_errors[error_as_string] = 1

                overall_time += ms

                if args.verbose:
                    if resp is None:
                        print(
                            "%s  ???  ---- Connection Error!"
                            % helper.create_id(
                                first_id,
                                args.prefix,
                                args.postfix,
                                max_id_length,
                                args.protocol == helper.PROTOCOL_NGSI_LD,
                            )
                        )
                    else:
                        message = " ".join(resp.text.split())[0:120]
                        print(
                            "%s  %3i  %4i %s"
                            % (
                                helper.create_id(
                                    first_id,
                                    args.prefix,
                                    args.postfix,
                                    max_id_length,
                                    args.protocol == helper.PROTOCOL_NGSI_LD,
                                ),
                                resp.status_code,
                                ms,
                                message,
                            )
                        )

        if not args.static_id:
            thing_id = sensor_things.INVALID_ID
            first_id += 1

        if args.frequency is not None:
            delta = datetime.now() - start_time_ms
            milliseconds = int(delta.total_seconds() * 1000)
            sleep_for_ms = args.frequency - milliseconds
            sleep_for_ms -= 10  # give some extra for the call itself
            if sleep_for_ms > 0:
                sleep(sleep_for_ms / 1000)


def signal_handler(*_):
    stop_send_threads()
    print("\nInterrupted!")
    output.show_result(False, start, errors, unique_errors)
    sys.exit(0)


def signal_handler_delete(*_):
    stop_delete_thread()
    print("\nInterrupted!")
    output.show_result(True, start, errors, unique_errors)
    sys.exit(0)


def wait_for_delete_thread(args):
    ready = False
    while not ready:
        sleep(0.5)
        if overall_messages > 0 and not args.verbose:
            if args.unlimited:
                print(
                    "\rMessages sent: %i with %i not found and %i other error(s)      "
                    % (overall_messages, not_deleted, errors),
                    end="",
                    flush=True,
                )
            else:
                print(
                    "\rMessages sent: %i (%i%%) with %i not found and %i other error(s)      "
                    % (
                        overall_messages,
                        (overall_messages / (args.delete[1] - args.delete[0] + 1))
                        * 100,
                        not_deleted,
                        errors,
                    ),
                    end="",
                    flush=True,
                )

        ready = True
        if delete_thread.is_alive():
            ready = False
            continue


def create_and_start_delete_thread(session, lock, args, max_id_length):
    global delete_thread

    delete_thread = Thread(target=do_delete, args=(session, lock, args, max_id_length))

    delete_thread.daemon = True
    delete_thread.start()


def stop_delete_thread():
    global delete_thread, halt

    halt = True
    while delete_thread.is_alive():
        sleep(0.5)


def handle_delete(args, session, lock, max_id_length):
    global start, delete_thread

    # noinspection PyTypeChecker
    signal.signal(signal.SIGINT, signal_handler_delete)

    # print what will be done...
    output.print_server_used(True, args.server)
    output.print_type_of_server(args.protocol)
    output.print_will_delete(args)

    # dry run only?
    if args.dry_run:
        print("Dry run only. Exiting...", flush=True)
        sys.exit(0)

    start = datetime.now()

    if args.verbose:
        print("ID, Response-Code, Content")

    create_and_start_delete_thread(session, lock, args, max_id_length)

    wait_for_delete_thread(args)

    print("\nReady", flush=True)


def create_send_threads(args, mqtt_client, session, lock, max_id_length):
    print("Starting %i thread(s)" % args.num_threads, end="", flush=True)

    offset = 0

    for i in range(args.num_threads):
        t = Thread(
            target=do_send,
            args=(mqtt_client, session, lock, args, offset, max_id_length),
        )
        send_threads.append(t)
        print(".", end="", flush=True)

        if not args.static_id:
            offset += args.num_messages


def start_send_threads():
    for t in send_threads:
        t.daemon = True
        t.start()


def wait_for_send_threads(limit_time, verbose, msg_num, unlimited):
    ready = False
    temp_ms = 0

    while not ready:
        sleep(0.5)
        temp_end = datetime.now()
        temp_delta = temp_end - start
        temp_s = int(temp_delta.total_seconds())
        temp_ms = int(temp_delta.total_seconds() * 1000)

        if limit_time is not None and temp_s >= limit_time:
            ready = True
            continue

        if not verbose:
            output.print_messages_send(
                overall_messages, errors, overall_time, temp_ms, msg_num, unlimited
            )

        ready = True
        for t in send_threads:
            if t.is_alive():
                ready = False
                continue

    if not verbose:
        output.print_messages_send(
            overall_messages, errors, overall_time, temp_ms, msg_num, unlimited
        )

    if limit_time is not None:
        print("\nTime is up!", flush=True)

    return temp_ms


def stop_send_threads():
    global send_threads, halt

    halt = True

    ready = False
    while not ready:
        ready = True
        for t in send_threads:
            if t.is_alive():
                ready = False
                continue
        sleep(0.5)


def handle_send(args, session, lock, msg_num, max_id_length):
    global start, send_threads

    # noinspection PyTypeChecker
    signal.signal(signal.SIGINT, signal_handler)

    # print what will be done...
    output.print_server_used(False, args.server)
    output.print_type_of_server(args.protocol)
    output.print_schema(args)
    if args.datastream_id is None:
        output.print_id_used(args, msg_num)
    else:
        output.print_data_stream_id_used(args.datastream_id)
    if args.dry_run:
        output.print_payload(args)
    output.print_will_send_messages(args, msg_num)
    if args.frequency is not None:
        output.print_frequency(args.frequency, args.num_threads > 1)

    # dry run only?
    if args.dry_run:
        print("Dry run only. Exiting...", flush=True)
        sys.exit(0)

    mqtt_client = (
        sensor_things.init_mqtt(helper.create_host_url(args.server))
        if args.protocol == helper.PROTOCOL_SENSOR_THINGS_MQTT
        else None
    )

    create_send_threads(args, mqtt_client, session, lock, max_id_length)

    start = datetime.now()
    start_send_threads()

    print("Ready\nRunning...", flush=True)

    if args.verbose:
        print("ID, Response-Code, Response-Time, Content", flush=True)

    # Wait for all threads to finish
    ms = wait_for_send_threads(args.limit_time, args.verbose, msg_num, args.unlimited)

    if not args.verbose:
        print("", flush=True)
    print("Ready", flush=True)

    if mqtt_client is not None:
        mqtt_client.loop_stop()

    if args.verbose:
        output.print_messages_send(
            overall_messages, errors, overall_time, ms, msg_num, args.unlimited
        )
        print("")


def main(args=None):
    output.print_version()
    args = arguments.parse_arguments()
    # if there was any error on the arguments, the function already gave some hint and exited.

    # calculate the overall number of messages to be sent
    msg_num = args.num_threads * args.num_messages

    max_id_length = helper.calculate_max_id_length(args, msg_num)

    if args.delete:
        handle_delete(args, requests.Session(), Lock(), max_id_length)
    else:
        handle_send(args, requests.Session(), Lock(), msg_num, max_id_length)

    output.show_result(args.delete, start, errors, unique_errors)

    sys.exit(0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import json
import random
from datetime import datetime

# some "consts"
VERSION = "1.1.1"
CONTENT_TYPE = "Content-Type"
APPLICATION_JSON = "application/json"
# TODO: swagger says, content-type is "application/json;application/ld+json"!
APPLICATION_JSON_LD = "application/ld+json"
LD_PREFIX = "urn:"
PROTOCOL_NGSI_V2 = "NGSI-V2"
PROTOCOL_NGSI_LD = "NGSI-LD"
PROTOCOL_SENSOR_THINGS_HTTP = "SensorThings-HTTP"
PROTOCOL_SENSOR_THINGS_MQTT = "SensorThings-MQTT"


def get_version():
    return VERSION


def create_id(first_id, prefix, postfix, align, is_ngsi_ld):
    ret = str(first_id)

    if prefix is not None:
        ret = prefix + ret

    if postfix is not None:
        ret += postfix

    if is_ngsi_ld:
        ret = LD_PREFIX + ret

    if align == 0:
        return ret

    return ret.rjust(align, " ")


def create_host_url(host):
    if host.startswith("http://") or host.startswith("https://"):
        return host

    return "https://" + host


def create_value_from_attribute_args(attribute_args):
    value = None
    if attribute_args[1] == "i":
        f = int(attribute_args[2])
        value = f
        if len(attribute_args) > 3:
            t = int(attribute_args[3])
            value = random.randint(f, t)
    elif attribute_args[1] == "f":
        f = float(attribute_args[2])
        value = f
        if len(attribute_args) > 3:
            t = float(attribute_args[3])
            value = round(random.uniform(f, t), 1)

    return value


def create_thing_payload(thing_name, indent):
    payload = dict()

    payload["name"] = thing_name
    payload["description"] = "This Thing was created by opensim.py"

    location = dict()
    location["name"] = thing_name
    location["description"] = "This Location was created by opensim.py"
    location["encodingType"] = "application/vnd.geo+json"
    loc = dict()
    loc["type"] = "Point"
    coordinates = list()
    coordinates.append(8.759449427022547)
    coordinates.append(51.72663719884235)
    loc["coordinates"] = coordinates
    location["location"] = loc
    locations = list()
    locations.append(location)
    payload["Locations"] = locations

    if indent > 0:
        return json.dumps(payload, indent=indent)
    else:
        return json.dumps(payload)


def create_data_stream_payload(data_stream_name, indent):
    payload = dict()

    payload["name"] = data_stream_name
    payload["description"] = "This Datastream was created by opensim.py"
    payload[
        "observationType"
    ] = "http://www.opengis.net/def/observationType/OGC-OM/2.0/OM_Measurement"

    unit_of_measurement = dict()
    unit_of_measurement["name"] = "number"
    unit_of_measurement["symbol"] = "#"
    unit_of_measurement["definition"] = "unknown"

    payload["unitOfMeasurement"] = unit_of_measurement

    observed_property = dict()
    observed_property["name"] = data_stream_name
    observed_property["description"] = "This ObservedProperty was created by opensim.py"
    observed_property["definition"] = "unknown"

    payload["ObservedProperty"] = observed_property

    sensor = dict()
    sensor["name"] = "Pseudo Sensor created by opensim.py"
    sensor["description"] = "I'm not real!"
    sensor["encodingType"] = "application/pdf"
    sensor["metadata"] = "https://i-dont-have.any/data-sheet.pdf"

    payload["Sensor"] = sensor

    if indent > 0:
        return json.dumps(payload, indent=indent)
    else:
        return json.dumps(payload)


def create_observation_payload(value, indent):
    payload = dict()

    time_string = datetime.now().strftime("%Y-%m-%dT%H:%M:%S.00Z")

    payload["result"] = value
    payload["phenomenonTime"] = time_string
    payload["resultTime"] = time_string

    if indent > 0:
        return json.dumps(payload, indent=indent)
    else:
        return json.dumps(payload)


def create_payload_ngsi_v2(first_id, meta_data, args):
    payload = dict()
    if meta_data:
        if first_id is not None:
            payload["id"] = create_id(
                first_id,
                args.prefix,
                args.postfix,
                0,
                args.protocol == PROTOCOL_NGSI_LD,
            )
        if args.type is not None:
            payload["type"] = args.type

    if args.date_times is not None:
        for date_time in args.date_times:
            attr = {
                "type": "DateTime",
                "value": datetime.now().strftime("%Y-%m-%dT%H:%M:%S.00Z"),
            }
            payload[date_time[0]] = attr

    if args.numbers is not None:
        for number in args.numbers:
            attribute_args = number[0].split(",")

            attr = {}

            # type
            if attribute_args[1] == "i" or attribute_args[1] == "f":
                attr["type"] = "Number"

            # value
            value = create_value_from_attribute_args(attribute_args)
            attr["value"] = value
            payload[attribute_args[0]] = attr

    if args.strings is not None:
        for string in args.strings:
            attr = {"type": "Text", "value": string[1]}
            payload[string[0]] = attr

    if args.booleans is not None:
        for boolean in args.booleans:
            attr = {"type": "Boolean"}

            # value
            if boolean[1] == "false":
                attr["value"] = False
            elif boolean[1] == "true":
                attr["value"] = True
            else:
                attr["value"] = bool(random.getrandbits(1))
            payload[boolean[0]] = attr

    if args.locations is not None:
        for location in args.locations:
            attribute_args = location[0].split(",")

            attr = {"type": "geo:json"}

            # value
            coord = {"type": "Point"}
            lat_from = float(attribute_args[1])
            long_from = float(attribute_args[2])
            coord["coordinates"] = [lat_from, long_from]

            if len(attribute_args) > 3:
                lat_to = float(attribute_args[3])
                long_to = float(attribute_args[4])
                coord["coordinates"] = [
                    random.uniform(lat_from, lat_to),
                    random.uniform(long_from, long_to),
                ]

            attr["value"] = coord
            payload[attribute_args[0]] = attr

    if args.indent > 0:
        return json.dumps(payload, indent=args.indent)
    else:
        return json.dumps(payload)


def create_payload_ngsi_ld(first_id, args, is_post):
    payload = dict()

    if first_id is not None:
        context = {
            "id": create_id(
                first_id,
                args.prefix,
                args.postfix,
                0,
                args.protocol == PROTOCOL_NGSI_LD,
            ),
            "type": args.type,
        }
    else:
        context = {"type": args.type}
    payload["@context"] = context

    if is_post:
        if first_id is not None:
            payload["id"] = create_id(
                first_id,
                args.prefix,
                args.postfix,
                0,
                args.protocol == PROTOCOL_NGSI_LD,
            )
        payload["type"] = args.type

    if args.date_times is not None:
        for date_time in args.date_times:
            attr = {
                "type": "DateTime",
                "value": datetime.now().strftime("%Y-%m-%dT%H:%M:%S.00Z"),
            }
            payload[date_time[0]] = attr

    if args.numbers is not None:
        for number in args.numbers:
            attribute_args = number[0].split(",")

            attr = {}

            # type
            if attribute_args[1] == "i" or attribute_args[1] == "f":
                attr["type"] = "Property"  # TODO: "Number" must be accepted!

            # value
            value = create_value_from_attribute_args(attribute_args)
            attr["value"] = value
            payload[attribute_args[0]] = attr

    if args.strings is not None:
        for string in args.strings:
            attr = {"type": "Text", "value": string[1]}
            payload[string[0]] = attr

    if args.booleans is not None:
        for boolean in args.booleans:
            attr = {"type": "Boolean"}

            # value
            if boolean[1] == "false":
                attr["value"] = False
            elif boolean[1] == "true":
                attr["value"] = True
            else:
                attr["value"] = bool(random.getrandbits(1))
            payload[boolean[0]] = attr

    if args.locations is not None:
        for location in args.locations:
            attribute_args = location[0].split(",")

            attr = {"type": "geo:json"}

            # value
            coord = {"type": "Point"}
            lat_from = float(attribute_args[1])
            long_from = float(attribute_args[2])
            coord["coordinates"] = [lat_from, long_from]

            if len(attribute_args) > 3:
                lat_to = float(attribute_args[3])
                long_to = float(attribute_args[4])
                coord["coordinates"] = [
                    random.uniform(lat_from, lat_to),
                    random.uniform(long_from, long_to),
                ]

            attr["value"] = coord
            payload[attribute_args[0]] = attr

    if args.indent > 0:
        return json.dumps(payload, indent=args.indent)
    else:
        return json.dumps(payload)


def calculate_max_id_length(args, msg_num):
    if args.delete:
        max_id_length = len(str(args.delete[1]))
        if args.prefix is not None:
            max_id_length += len(args.prefix)
        if args.postfix is not None:
            max_id_length += len(args.postfix)
    else:
        if args.static_id:
            max_id_length = len(str(args.first_id))
        else:
            max_id_length = len(str(args.first_id + msg_num - 1))

        if args.prefix is not None:
            max_id_length += len(args.prefix)
        if args.postfix is not None:
            max_id_length += len(args.postfix)

    return max_id_length



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from datetime import datetime

from . import helper


def print_version():
    print(
        "Open Smart City-Sim, Copyright (c) 2021 Will Freitag, Version %s"
        % helper.get_version(),
        flush=True,
    )


def print_delimiter():
    print(
        "----------------------------------------------"
        "-------------------------------------------",
        flush=True,
    )


def print_duration(what, start, end):
    print(
        "%s ran from [%s] to [%s]"
        % (
            what,
            start.strftime("%d.%m.%Y - %H:%M:%S"),
            end.strftime("%d.%m.%Y - %H:%M:%S"),
        )
    )


def print_time_spent(ms):
    if ms < 1000:
        print("Time spent: %i ms" % ms, flush=True)
    elif ms < 60000:
        print("Time spent: %i sec" % (ms / 1000), flush=True)
    else:
        minutes = ms / 60000
        ms -= int(minutes) * 60000
        print("Time spent: %i:%02i" % (minutes, ms / 1000), flush=True)


def show_result(delete, start, errors, unique_errors):
    if errors > 0:
        print("\nErrors:", flush=True)
        for key in unique_errors.keys():
            print('%i time(s) "%s"' % (unique_errors[key], key))

    end = datetime.now()
    delta = end - start
    ms = int(delta.total_seconds() * 1000)

    print_delimiter()
    print_time_spent(ms)

    if delete:
        print_duration("Delete", start, end)
    else:
        print_duration("Test", start, end)


def print_server_used(delete, host):
    if delete:
        print("Will delete on %s" % helper.create_host_url(host), flush=True)
    else:
        print("Running load against %s" % helper.create_host_url(host), flush=True)


def print_type_of_server(protocol):
    print("Type of server is: %s" % protocol, flush=True)


def print_schema(args):
    if args.protocol == helper.PROTOCOL_NGSI_V2:
        if args.insert_always:
            print(
                "Note: POST-always schema is used to store contexts. "
                "(Use --help to get more information)",
                flush=True,
            )
        else:
            print(
                "Note: PATCH/POST schema is used to store contexts. "
                "(Use --help to get more information)",
                flush=True,
            )


def print_data_stream_id_used(datastream_id):
    print(
        "The Datastream-Id %i will be used for ALL Observations! "
        "(Use --help to get more information)" % datastream_id,
        flush=True,
    )


def print_id_used(args, msg_num):
    if args.static_id:
        print(
            'ID "%s" will be used for all messages.'
            % helper.create_id(
                args.first_id,
                args.prefix,
                args.postfix,
                0,
                args.protocol == helper.PROTOCOL_NGSI_LD,
            ),
            flush=True,
        )
    else:
        if msg_num == 1:
            print(
                'The used ID will be "%s".'
                % helper.create_id(
                    args.first_id,
                    args.prefix,
                    args.postfix,
                    0,
                    args.protocol == helper.PROTOCOL_NGSI_LD,
                ),
                flush=True,
            )
        elif args.unlimited:
            print(
                'The used ID will be starting at "%s" and increase continuously.'
                % helper.create_id(
                    args.first_id,
                    args.prefix,
                    args.postfix,
                    0,
                    args.protocol == helper.PROTOCOL_NGSI_LD,
                ),
                flush=True,
            )
        else:
            print(
                'The used ID will be from "%s" to "%s".'
                % (
                    helper.create_id(
                        args.first_id,
                        args.prefix,
                        args.postfix,
                        0,
                        args.protocol == helper.PROTOCOL_NGSI_LD,
                    ),
                    helper.create_id(
                        args.first_id + msg_num - 1,
                        args.prefix,
                        args.postfix,
                        0,
                        args.protocol == helper.PROTOCOL_NGSI_LD,
                    ),
                ),
                flush=True,
            )


def print_payload(args):
    if args.protocol == helper.PROTOCOL_NGSI_V2:
        if args.insert_always:
            print(
                "The payload will look like:\n%s"
                % helper.create_payload_ngsi_v2(args.first_id, True, args),
                flush=True,
            )
        else:
            print(
                "The payload will look like:\n%s"
                % helper.create_payload_ngsi_v2(None, False, args),
                flush=True,
            )
    elif args.protocol == helper.PROTOCOL_NGSI_LD:
        print(
            "The payload will look like:\n%s"
            % helper.create_payload_ngsi_ld(args.first_id, args, True),
            flush=True,
        )
    else:
        thing_name = helper.create_id(
            args.first_id,
            args.prefix,
            args.postfix,
            0,
            args.protocol == helper.PROTOCOL_NGSI_LD,
        )

        print(
            "The payload will look like:\nThing:\n%s\n\nDatastream: \n%s\n\nObservation: \n%s"
            % (
                helper.create_thing_payload(thing_name, args.indent),
                helper.create_data_stream_payload("AttributeName", args.indent),
                helper.create_observation_payload(1.5, args.indent),
            ),
            flush=True,
        )


def print_will_delete(args):
    print(
        'Will delete %i contexts from "%s" to "%s".'
        % (
            args.delete[1] - args.delete[0] + 1,
            helper.create_id(
                args.delete[0],
                args.prefix,
                args.postfix,
                0,
                args.protocol == helper.PROTOCOL_NGSI_LD,
            ),
            helper.create_id(
                args.delete[1],
                args.prefix,
                args.postfix,
                0,
                args.protocol == helper.PROTOCOL_NGSI_LD,
            ),
        ),
        flush=True,
    )


def print_will_send_messages(args, msg_num):
    if args.unlimited:
        if args.limit_time is not None:
            print(
                "Will send messages in %i thread(s) for %i seconds."
                % (args.num_threads, args.limit_time),
                flush=True,
            )
        else:
            print(
                "Will send infinite messages in %i thread(s). Hit 'Ctrl-C' to interrupt."
                % args.num_threads,
                flush=True,
            )
    else:
        if msg_num == 1:
            print("Will send one message.", flush=True)
        else:
            print(
                "Will send %i messages in %i thread(s)." % (msg_num, args.num_threads),
                flush=True,
            )


def print_frequency(frequency, more_than_one_thread):
    if more_than_one_thread:
        print(
            "The frequency of messages sent is limited to %i milliseconds (per thread!)."
            % frequency,
            flush=True,
        )
    else:
        print(
            "The frequency of messages sent is limited to %i milliseconds." % frequency,
            flush=True,
        )


def print_messages_send(overall_messages, errors, overall_time, ms, msg_num, unlimited):
    net_messages = overall_messages - errors
    msg_per_second = 0
    response_time = -1

    if net_messages > 0:
        response_time = int(overall_time / net_messages)

    if ms > 0:
        msg_per_second = max(0, int(((overall_messages - errors) * 1000) / ms))

    if unlimited:
        # In this case, there is no percentage
        msg = (
            "\rMessages sent: %i with %i error(s), "
            "avg. response-time: %s ms, load: %i msg/sec      "
            % (
                overall_messages,
                errors,
                str(response_time) if response_time >= 0 else "--",
                msg_per_second,
            )
        )
    else:
        # Here we go with some percentage
        percentage = (overall_messages / msg_num) * 100
        msg = (
            "\rMessages sent: %i (%i%%) with %i error(s), "
            "avg. response-time: %s ms, load: %i msg/sec      "
            % (
                overall_messages,
                percentage,
                errors,
                str(response_time) if response_time >= 0 else "--",
                msg_per_second,
            )
        )

    print(msg, end="", flush=True)



import os

from setuptools import find_packages, setup

folder = os.path.dirname(os.path.realpath(__file__))
requirementPath = folder + '/requirements.txt'
install_requires = []
if os.path.isfile(requirementPath):
    with open(requirementPath) as f:
        install_requires = f.read().splitlines()

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("CHANGELOG.md", "r", encoding="utf-8") as fh:
    change_log = fh.read()

setup(
    name="oscsim",
    version="1.1.1",
    author="Will Freitag",
    author_email="Wilhelm.Freitag@omp.de",
    description="Open Smart City-Sim is a lightweight tool to send test data "
                "to an Orion Context Broker or FROST-Server respectively.",
    license='MIT License',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/HYPERTEGRITY-AG/OpenSim",
    project_urls={
        "Bug Tracker": "https://github.com/HYPERTEGRITY-AG/OpenSim/projects/1",
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    package_dir={'': 'src'},
    packages=find_packages('src'),
    entry_points={'console_scripts': ['oscsim = oscsim.run:main']},
    python_requires='>=3.6',
    install_requires=install_requires
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import argparse
import shutil
import textwrap as _textwrap

from . import helper


def parse_arguments():
    class MultiLineFormatter(argparse.HelpFormatter):
        def __init__(self, prog, indent_increment=2, max_help_position=24):
            argparse.HelpFormatter.__init__(
                self,
                prog,
                indent_increment,
                max_help_position,
                width=shutil.get_terminal_size().columns,
            )

        def _fill_text(self, text, width, indent):
            paragraphs = text.split("|n")
            multi_line_text = ""
            for paragraph in paragraphs:
                formatted_paragraph = (
                    _textwrap.fill(
                        paragraph,
                        width,
                        initial_indent=indent,
                        subsequent_indent=indent,
                    )
                    + "\n"
                )
                multi_line_text = multi_line_text + formatted_paragraph
            return multi_line_text

    # noinspection PyTypeChecker
    parser = argparse.ArgumentParser(
        description="Tool to create some load on Orion Context Broker/FROST-Server.",
        formatter_class=MultiLineFormatter,
        allow_abbrev=False,
    )

    parser.epilog = (
        "Example #1:|n"
        "oscsim -s my-host.com -H Authorization "
        "'Bearer 039ea6d72a2f32227c2110bd8d78aae33acd6782' "
        "-H Fiware-service curltest|n"
        "One message is sent using id '1'.|n"
        "The tenant 'curltest' will be used "
        "as 'Fiware-service' in the "
        "header of the post.|n"
    )

    parser.epilog += (
        "|n|nExample #2:|n"
        "oscsim -s my-host.com -n 2 -m 50 ...|n"
        "100 messages will be sent (2 threads are sending 50 messages each)."
        "The id will be looped from '1' to '100'.|n"
    )

    parser.epilog += (
        "|n|nExample #3:|n"
        "oscsim -s my-host.com -n 5 -m 100 -f 123 -c ...|n"
        "500 messages will be sent (5 threads are sending 100 messages each)."
        "The id '123' (-f is first id) will be used for all messages (-c is static id).|n"
    )

    parser.epilog += (
        "|n|nExample #4:|n"
        "The payload that will be sent is constructed from the -y and "
        "the -aX parameters. Example:|n"
        "oscsim -y WeatherObserved -an temperature,f,-20,50 "
        "-an precipitation,i,1,20 ...|n"
        "will generate a payload looking like:|n"
        "{|n"
        '  "id":"1",|n'
        '  "type":"WeatherObserved",|n'
        '  "temperature": {|n'
        '    "type": "Number",|n'
        '    "value": -2.3|n'
        "  },|n"
        '  "precipitation": {|n'
        '    "type": "Number",|n'
        '    "value": 13|n'
        "  },|n"
        "}|n"
    )

    parser.epilog += (
        "|n|nExample #5:|n"
        "oscsim -d 100 200 -s my-host.com "
        "-H Authorization 'Bearer 039ea6d72a2f32227c2110bd8d78aae33acd6782'|n"
        "|n"
        "This will delete all IDs starting from 100 to 200 (inclusive)."
    )

    parser.add_argument(
        "-s",
        "--server",
        metavar="[protocol]host-name",
        dest="server",
        help='This host-name will be prepended by "https://", if protocol '
        "is omitted and appended "
        'with "/v2/" (NGSI-V2), "/ngsi-ld/v1" '
        '(NGSI-LD) or "/v1.1/" (SensorThings) resp. '
        "depending on the server-type (see -p/--protocol).",
        required=True,
    )

    parser.add_argument(
        "-p",
        "--protocol",
        choices=[
            helper.PROTOCOL_NGSI_V2,
            helper.PROTOCOL_NGSI_LD,
            helper.PROTOCOL_SENSOR_THINGS_MQTT,
            helper.PROTOCOL_SENSOR_THINGS_HTTP,
        ],
        default=helper.PROTOCOL_NGSI_V2,
        dest="protocol",
        help="Define the type of server. [Default: %s]" % helper.PROTOCOL_NGSI_V2,
    )

    parser.add_argument(
        "-i",
        "--insert-always",
        dest="insert_always",
        action="store_true",
        default=False,
        help="[Only NGSI-V2 and NGSI-LD!] If set, the contexts will always be inserted "
        "(via POST with option 'upsert') instead "
        "of trying to update first (via PATCH) and insert (via POST), if not "
        "existing (i.e. PATCH returns '404 Not Found').",
    )

    parser.add_argument(
        "-a",
        "--datastream-id",
        metavar="id",
        dest="datastream_id",
        help="[Only SensorThings!] If set, this Datastream-Id will be used for ALL "
        "Observations, instead of first searching for the Thing by it's name "
        "and the correct "
        "Datastream-Id afterwards.",
        type=int,
    )

    parser.add_argument(
        "-H",
        "--header",
        metavar=("key", "value"),
        dest="headers",
        action="append",
        nargs=2,
        help="Define a header by key and value.",
    )

    parser.add_argument(
        "-f",
        "--first-id",
        metavar="id",
        dest="first_id",
        help="Define the first id to be used or the one to be used "
        "if '-c/--static-id' is "
        "set. [Default: 1]",
        default=1,
        type=int,
    )

    parser.add_argument(
        "-e",
        "--prefix",
        dest="prefix",
        help="If set, the prefix will be prepended to the generated id.",
    )

    parser.add_argument(
        "-o",
        "--postfix",
        dest="postfix",
        help="If set, the postfix will be appended to the generated id.",
    )

    parser.add_argument(
        "-c",
        "--static-id",
        dest="static_id",
        action="store_true",
        default=False,
        help="If set, the id will not increment (i.e. -n times -m messages will be "
        "sent with the same id ['-f/--first-id' or '1' if omitted]).",
    )

    parser.add_argument(
        "-n",
        "--num-threads",
        metavar="num",
        dest="num_threads",
        help="Define, how many threads shall be used. [Default: 1]",
        default=1,
        type=int,
    )

    parser.add_argument(
        "-m",
        "--messages",
        metavar="num",
        dest="num_messages",
        help="Define, how many messages per thread shall be "
        "sent (ignored, if '-u/--unlimited' "
        "ist set). [Default: 1]",
        default=1,
        type=int,
    )

    parser.add_argument(
        "-u",
        "--unlimited",
        dest="unlimited",
        action="store_true",
        default=False,
        help="If set, '-m/--messages' is ignored and infinite messages "
        "will be send "
        "(in '-n/--num-threads' threads). "
        " Hit 'Ctrl-C' to interrupt or set '-l/--limit-time'.",
    )

    parser.add_argument(
        "-q",
        "--frequency",
        metavar="milliseconds",
        dest="frequency",
        help="If set, limits the frequency of the messages sent to the given "
        "number (per thread!).",
        type=int,
    )

    parser.add_argument(
        "-l",
        "--limit-time",
        metavar="seconds",
        dest="limit_time",
        help="Only in conjunction with '-u/--unlimited': Stops after the given "
        "time in seconds.",
        type=int,
    )

    parser.add_argument(
        "-y",
        "--type",
        metavar="name",
        dest="type",
        help="[Only NGSI-V2 and NGSI-LD!] If set, this type-name will be used in the payload.",
    )

    parser.add_argument(
        "-an",
        "--attribute-number",
        metavar="name,type,number[,max-number]",
        dest="numbers",
        action="append",
        nargs=1,
        help="Define a number attribute used for the payload by 'name' (The name "
        "of the "
        "attribute, e.g.: temperature), 'type' (One of i [integer] or "
        "f [floating point])"
        "and 'number' (The value to be used). If 'max-number' is set, the "
        "number written will be"
        " randomly between 'number' and 'max-number' (each including). "
        "Note: Multiple "
        "number attributes can be defined by repeating -an.",
    )

    parser.add_argument(
        "-as",
        "--attribute-string",
        metavar=("name", "value"),
        dest="strings",
        action="append",
        nargs=2,
        help="Define a string attribute used for the payload by 'name' "
        "(The name of the attribute, e.g.: instruction)"
        " and 'value' (the actual string). Note: Multiple "
        "string attributes can be defined by repeating -as.",
    )

    parser.add_argument(
        "-ad",
        "--attribute-date",
        metavar="name",
        dest="date_times",
        action="append",
        nargs=1,
        help="[Only NGSI-V2!] Define a DateTime attribute used for the payload "
        "by 'name' "
        "(The name of the attribute, "
        "e.g.: dateObserved). Note: The current time is used as value. "
        "Multiple DateTime "
        "attributes can be defined by repeating -ad.",
    )

    parser.add_argument(
        "-al",
        "--attribute-location",
        metavar="name,lat,long[,max-lat,max-long]",
        dest="locations",
        action="append",
        nargs=1,
        help="[Only NGSI-V2!] Define a location attribute used for the payload "
        "by 'name' (The name "
        "of the attribute, e.g.: position), 'lat' (The value for latitude) "
        "and 'long' "
        "(The value for longitude). If 'max-lat' and 'max-long' are set, "
        "the location "
        "written will be randomly between 'lat' and 'max-lat' and 'long' "
        "and 'max-long' "
        "resp. (each including). Note: Multiple location attributes "
        "can be defined by repeating -al.",
    )

    parser.add_argument(
        "-ab",
        "--attribute-boolean",
        metavar=("name", "value"),
        dest="booleans",
        action="append",
        nargs=2,
        help="[Only NGSI-V2!] Define a boolean attribute used for the payload "
        "by 'name' "
        "(The name of the attribute, "
        "e.g.: public) and 'value' (One of 'true', 'false' or 'toggle' "
        "[ie. randomly switch "
        "between true and false]). Note: Multiple boolean attributes can "
        "be defined "
        "by repeating -ab.",
    )

    parser.add_argument(
        "-ai",
        "--attribute-indent",
        metavar="indent",
        dest="indent",
        help="Define the number of characters for indenting the created payload. "
        "[Default: 0]",
        default=0,
        type=int,
    )

    parser.add_argument(
        "-r",
        "--dry-run",
        dest="dry_run",
        action="store_true",
        default=False,
        help="Do a dry run only - giving the chance to review what WOULD be done "
        "incl. seeing what the "
        "payload will look like.",
    )

    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbose",
        action="store_true",
        default=False,
        help="Generate verbose output.",
    )

    parser.add_argument(
        "-d",
        "--delete",
        metavar=("from", "to"),
        dest="delete",
        type=int,
        nargs=2,
        help="If set, the entities within the given range "
        '(including "from" and "to") will be deleted.',
    )

    args = parser.parse_args()

    check_arguments(parser, args)

    return args


def check_arguments(parser, args):
    if args.delete is not None:
        if args.delete[0] > args.delete[1]:
            # switch the indexes
            t = args.delete[0]
            args.delete[0] = args.delete[1]
            args.delete[1] = t
        if args.unlimited:
            parser.error(
                "The switch '-u/--unlimited' cannot be used when deleting messages "
                "[-d/--delete]!"
            )
    else:
        if args.datastream_id is not None and args.datastream_id < 0:
            parser.error(
                "Please define a positive number for Datastream-Id [-a/--datastream-id]"
            )

        if args.num_threads == 0:
            parser.error(
                "Without any thread, no messages will be sent at all! "
                "[-n/--num-threads > 0]"
            )

        if args.num_messages <= 0:
            parser.error(
                "Please consider increasing the number of messages! [-m/--messages > 0]"
            )

        if args.limit_time is not None and not args.unlimited:
            parser.error(
                "Limiting the time [-l/--limit-time] is only valid with infinite "
                "run! [-u/--unlimited]"
            )

        if args.limit_time is not None and args.limit_time <= 0:
            parser.error("Please define a positive number for limit [-l/--limit-time]")

        if args.frequency is not None and args.frequency <= 0:
            parser.error(
                "Please define a positive number for frequency [-q/--frequency]"
            )

        if args.numbers is not None:
            for number in args.numbers:
                attribute_args = number[0].split(",")
                if len(attribute_args) < 3 or len(attribute_args) > 4:
                    parser.error(
                        "-an argument ['-an %s'] expects 3 or 4 comma-delimited "
                        "parameters!" % ",".join(attribute_args)
                    )
                if attribute_args[1] == "i":
                    t = 0
                    f = 0
                    try:
                        f = int(attribute_args[2])
                    except ValueError:
                        parser.error(
                            'Please check attribute "%s": "number" must be an '
                            "integer!" % attribute_args[0]
                        )
                    if len(attribute_args) == 4:
                        try:
                            t = int(attribute_args[3])
                        except ValueError:
                            parser.error(
                                'Please check attribute "%s": "max-number" must '
                                "be an integer!" % attribute_args[0]
                            )
                        if t < f:
                            parser.error(
                                'Please check attribute "%s": "max-number" must '
                                "be greater "
                                'than or equal to "number"!' % attribute_args[0]
                            )
                elif attribute_args[1] == "f":
                    t = 0.0
                    f = 0.0
                    try:
                        f = float(attribute_args[2])
                    except ValueError:
                        parser.error(
                            'Please check attribute "%s": "number" must be a '
                            "floating point number!" % attribute_args[0]
                        )
                    if len(attribute_args) == 4:
                        try:
                            t = float(attribute_args[3])
                        except ValueError:
                            parser.error(
                                'Please check attribute "%s": "max-number" must be a '
                                "floating point number!" % attribute_args[0]
                            )
                        if t < f:
                            parser.error(
                                'Please check attribute "%s": "max-number" must be '
                                "greater "
                                'than or equal to "number"!' % attribute_args[0]
                            )
                else:
                    parser.error(
                        'Please check attribute "%s": type must be one of [i | f]!'
                        % attribute_args[0]
                    )

        if args.locations is not None:
            for location in args.locations:
                attribute_args = location[0].split(",")
                if len(attribute_args) != 3 and len(attribute_args) != 5:
                    parser.error(
                        "-al argument ['-al %s'] expects 3 or 5 comma-delimited "
                        "parameters!" % ",".join(attribute_args)
                    )
                lat_from = 0.0
                lat_to = 0.0
                long_from = 0.0
                long_to = 0.0
                try:
                    lat_from = float(attribute_args[1])
                except ValueError:
                    parser.error(
                        'Please check attribute "%s": "lat" must be a floating '
                        "point number!" % attribute_args[0]
                    )
                if lat_from < -90.0 or lat_from > 90.0:
                    parser.error(
                        'Please check attribute "%s": "lat" must be in a range from '
                        "-90.0 to 90.0!" % attribute_args[0]
                    )

                try:
                    long_from = float(attribute_args[2])
                except ValueError:
                    parser.error(
                        'Please check attribute "%s": "long" must be a floating '
                        "point number!" % attribute_args[0]
                    )
                if long_from < -180.0 or long_from > 180.0:
                    parser.error(
                        'Please check attribute "%s": "long" must be in a range from '
                        "-180.0 to 180.0!" % attribute_args[0]
                    )

                if len(attribute_args) == 5:
                    try:
                        lat_to = float(attribute_args[3])
                    except ValueError:
                        parser.error(
                            'Please check attribute "%s": "max-lat" must be a '
                            "floating "
                            "point number!" % attribute_args[0]
                        )

                    if lat_to < -90.0 or lat_to > 90.0:
                        parser.error(
                            'Please check attribute "%s": "max-lat" must be in a '
                            "range "
                            "from -90.0 to 90.0!" % attribute_args[0]
                        )

                    try:
                        long_to = float(attribute_args[4])
                    except ValueError:
                        parser.error(
                            'Please check attribute "%s": "max-long" must be a '
                            "floating point number!" % attribute_args[0]
                        )
                    if long_to < -180.0 or long_to > 180.0:
                        parser.error(
                            'Please check attribute "%s": "max-long" must be in a '
                            "range "
                            "from -180.0 to 180.0!" % attribute_args[0]
                        )

                    if lat_to < lat_from:
                        parser.error(
                            'Please check attribute "%s": "max-lat" must be greater '
                            "than "
                            'or equal to "lat"!' % attribute_args[0]
                        )

                    if long_to < long_from:
                        parser.error(
                            'Please check attribute "%s": "max-long" must be greater '
                            "than "
                            'or equal to "long"!' % attribute_args[0]
                        )

        if args.booleans is not None:
            for boolean in args.booleans:
                if (
                    boolean[1] != "true"
                    and boolean[1] != "false"
                    and boolean[1] != "toggle"
                ):
                    parser.error(
                        'Please check attribute "%s": "value" must be one '
                        'of "true", '
                        '"false" or "toggle"!' % boolean[0]
                    )

        if not args.delete:
            # check scheme?
            if args.insert_always and (
                args.protocol != helper.PROTOCOL_NGSI_V2
                and args.protocol != helper.PROTOCOL_NGSI_LD
            ):
                parser.error(
                    "Insert always scheme [-i/--insert-always] is only valid "
                    "for NGSI-V2 and NGSI-LD!"
                )

            # is there any payload?
            if (
                args.protocol == helper.PROTOCOL_NGSI_V2
                or args.protocol != helper.PROTOCOL_NGSI_LD
            ):
                if (
                    ((args.numbers is None) or (len(args.numbers) == 0))
                    and ((args.booleans is None) or (len(args.booleans) == 0))
                    and ((args.locations is None) or (len(args.locations) == 0))
                    and ((args.strings is None) or (len(args.strings) == 0))
                    and ((args.date_times is None) or (len(args.date_times) == 0))
                ):
                    parser.error("Please define any payload!")
            else:
                if ((args.numbers is None) or (len(args.numbers) == 0)) and (
                    (args.strings is None) or (len(args.strings) == 0)
                ):
                    parser.error("Please define any payload!")

    return True
